import os
import subprocess
from pydub import AudioSegment, silence
import csv

# Static variables - Set these before running the script
INPUT_FOLDER = r"F:\coqui_test\sounds"
OUTPUT_FOLDER = r"F:\coqui_test\training_dataset"
DURATION_LIMIT = 532 * 60 * 1000  # Duration limit in milliseconds (5 minutes by default)
MIN_CHUNK_LENGTH = 1500  # Minimum duration of a chunk in milliseconds 1500 = 1.5 second
MAX_CHUNK_LENGTH = 12000  # Maximum duration of a chunk in milliseconds 12000 = 12 second

# Parameters for silence detection and chunk splitting
MIN_SILENCE_LEN = 700  # Minimum length of silence to be used for splitting, in milliseconds
SILENCE_THRESH = -24   # Silence threshold (in dBFS) for detecting silence. Lower values mean more silence is detected
KEEP_SILENCE = 150     # Amount of silence to leave at the beginning and end of chunks, in milliseconds

# Ensure the output subfolder for WAV files exists
os.makedirs(os.path.join(OUTPUT_FOLDER, "wavs"), exist_ok=True)
print("Initialized and output folders ready.")

def extract_audio(input_file, output_file, duration_limit=None):
    print(f"Extracting audio from {input_file}...")
    if duration_limit:
        command = f"ffmpeg -i \"{input_file}\" -t {duration_limit / 1000} -ar 22050 -ac 1 \"{output_file}\""
    else:
        command = f"ffmpeg -i \"{input_file}\" -ar 22050 -ac 1 \"{output_file}\""
    subprocess.run(command, shell=True)
    print("Extraction complete.")

def split_and_process_chunks(chunks):
    processed_chunks = []
    buffer_chunk = None
    append_count = 0  # Initialize counter for append operations
    split_count = 0   # Initialize counter for splits
    new_sub_chunks_count = 0  # Initialize counter for new sub-chunks

    for chunk in chunks:
        if buffer_chunk:
            chunk = buffer_chunk + chunk
            buffer_chunk = None

        if MIN_CHUNK_LENGTH <= len(chunk) <= MAX_CHUNK_LENGTH:
            processed_chunks.append(chunk)
        elif len(chunk) < MIN_CHUNK_LENGTH:
            buffer_chunk = chunk
            append_count += 1
        else:
            split_count += 1  # Increment split counter
            # Split the chunk further
            sub_chunks = silence.split_on_silence(chunk, min_silence_len=int(MIN_SILENCE_LEN/2), silence_thresh=chunk.dBFS + SILENCE_THRESH, keep_silence=KEEP_SILENCE)
            new_sub_chunks_count += len(sub_chunks)  # Increment new sub-chunks counter
            for sub_chunk in sub_chunks:
                if MIN_CHUNK_LENGTH <= len(sub_chunk) <= MAX_CHUNK_LENGTH:
                    processed_chunks.append(sub_chunk)
                    append_count += 1

    if buffer_chunk:
        processed_chunks.append(buffer_chunk)
        append_count += 1

    return processed_chunks, append_count, split_count, new_sub_chunks_count

def split_audio(input_file):
    print(f"Splitting audio in {input_file}...")
    sound = AudioSegment.from_wav(input_file)
    all_chunks = silence.split_on_silence(sound, min_silence_len=MIN_SILENCE_LEN, silence_thresh=sound.dBFS + SILENCE_THRESH, keep_silence=KEEP_SILENCE)
    print(f"Main generated all_chunks count {len(all_chunks)}.")
    processed_chunks, append_count, split_count, new_sub_chunks_count = split_and_process_chunks(all_chunks)
    print(f"Splitting complete. {len(processed_chunks)} valid chunks identified.")
    print(f"'processed_chunks.append' called {append_count} times.")
    print(f"'Split the chunk further' called {split_count} times, generating {new_sub_chunks_count} new sub-chunks.")
    return processed_chunks

def process_files():
    file_counter = 100
    total_files = len([name for name in os.listdir(INPUT_FOLDER) if name.endswith((".mp3", ".wav", ".m4a", ".flac", ".mp4", ".mkv", ".avi"))])
    print(f"Total files to process: {total_files}")

    for filename in os.listdir(INPUT_FOLDER):
        if filename.endswith((".mp3", ".wav", ".m4a", ".flac", ".mp4", ".mkv", ".avi")):
            print(f"Processing file {file_counter} of {total_files}: {filename}")
            base_name = os.path.splitext(filename)[0]
            intermediate_wav = os.path.join(OUTPUT_FOLDER, f"temp_{base_name}.wav")
            extract_audio(os.path.join(INPUT_FOLDER, filename), intermediate_wav, DURATION_LIMIT)
            audio_chunks = split_audio(intermediate_wav)

            for i, chunk in enumerate(audio_chunks):
                chunk_file_name = f"sound-{file_counter}-{10000+i+1}.wav"
                chunk_name = os.path.join(OUTPUT_FOLDER, "wavs", chunk_file_name)
                chunk.export(chunk_name, format="wav")
                # Note: Transcription-related code removed here
                print(f"Chunk {i+1}/{len(audio_chunks)} of file {file_counter} processed.")

            file_counter += 1
            os.remove(intermediate_wav)
            print(f"Finished processing {filename}.\n")

    # Note: CSV writing code and Whisper model loading code removed here
    print("Data processing complete.")

# Run the script
process_files()
