from math import fabs
import os
import torch
import torchaudio
import datetime
import gradio as gr
import re
from TTS.tts.configs.xtts_config import XttsConfig
from TTS.tts.models.xtts import Xtts
from pedalboard import *
import noisereduce
from moviepy.editor import AudioFileClip, ColorClip
from seewav import visualize
import shutil

current_model = {
    'model_path': None,
    'model': None
}

enable_text_splitting=True

def split_text(text, max_char_length=250):
    sentences = re.split(r'(?<=[.!?])\s*|\n', text)
    chunks = []

    for sentence in sentences:
        stripped_sentence = sentence.strip()
        if len(stripped_sentence) > 0:
            if len(stripped_sentence) > max_char_length:
                chunks.extend(split_into_chunks(stripped_sentence, max_char_length))
            else:
                chunks.append(stripped_sentence)

    return chunks

def split_into_chunks(text, max_char_length):
    chunks = []
    current_chunk = ''

    for word in text.split():
        if len(current_chunk) + len(word) + 1 <= max_char_length:
            current_chunk += (' ' + word) if current_chunk else word
        else:
            chunks.append(current_chunk.strip())
            current_chunk = word

    if current_chunk:
        chunks.append(current_chunk.strip())

    return chunks

def load_model(model_path):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    print(f"Loading XTTS model on {device.upper()}...")

    if current_model['model_path'] == model_path:
        return current_model['model']

    config_file = os.path.join(os.path.dirname(model_path), "config.json")
    vocab_file = r"TTS_models\XTTS_v2.0_original_model_files\vocab.json"

    config = XttsConfig()
    config.load_json(config_file)
    model = Xtts.init_from_config(config)
    model.load_checkpoint(config, checkpoint_path=model_path, vocab_path=vocab_file, use_deepspeed=False)
    model.to(device)

    current_model['model_path'] = model_path
    current_model['model'] = model

    return model

def create_subfolder_for_date(folder_path):
    date_folder = datetime.datetime.now().strftime("%Y_%m_%d")
    subfolder_path = os.path.join(folder_path, date_folder)
    if not os.path.exists(subfolder_path):
        os.makedirs(subfolder_path)
    return subfolder_path

def inference(model_path, reference_audio_dropdown, reference_audio_upload, text, language, temperature, length_penalty, num_beams, repetition_penalty, top_k, top_p, do_sample, gpt_cond_len, gpt_cond_chunk_len, max_ref_len, sound_norm_refs, save_individual_audio, post_process_audio, save_as_mp4):
    print("Generating audio...")
    reference_audio = reference_audio_upload if reference_audio_upload else os.path.join(reference_audio_folder, reference_audio_dropdown)

    model = load_model(model_path)
    start_time = datetime.datetime.now()

    # Set parameters
    temperature = float(temperature)
    length_penalty = float(length_penalty)
    num_beams = int(num_beams)
    repetition_penalty = float(repetition_penalty)
    top_k = int(top_k)
    top_p = float(top_p)
    do_sample = bool(do_sample)
    gpt_cond_len = int(gpt_cond_len)
    gpt_cond_chunk_len = int(gpt_cond_chunk_len)
    max_ref_len = int(max_ref_len)
    sound_norm_refs = bool(sound_norm_refs)
    save_individual_audio = bool(save_individual_audio)
    post_process_audio = bool(post_process_audio) # New parameter
    save_as_mp4 = bool(save_as_mp4) # New parameter

    # Get conditioning latents
    gpt_cond_latent, speaker_embedding = model.get_conditioning_latents(
        audio_path=[reference_audio],
        gpt_cond_len=gpt_cond_len,
        gpt_cond_chunk_len=gpt_cond_chunk_len,
        max_ref_length=max_ref_len,
        sound_norm_refs=sound_norm_refs
    )
    timestamp = datetime.datetime.now().strftime("%H_%M_%S_%f")
    text_chunks = [text] if not save_individual_audio else split_text(text)
    enable_text_splitting = not save_individual_audio

    combined_audio = []
    counter = 1
    output_folder = "output_audios"
    subfolder_path = create_subfolder_for_date(output_folder)

    # Initialize audio processing board
    board = Pedalboard([
		NoiseGate(threshold_db=-30, ratio=1.5, release_ms=250),
		Compressor(threshold_db=12, ratio=2.5),
		LowShelfFilter(cutoff_frequency_hz=400, gain_db=5, q=1),
		HighpassFilter(cutoff_frequency_hz=80), 
		#Reverb(room_size=0.5, damping=0.5, wet_level=0.3, dry_level=0.7),  
		Limiter(threshold_db=-0.1),  
		Gain(gain_db=0)
	])

    for chunk in text_chunks:
        print(f"Generating audio {counter} / {len(text_chunks)}...")
        out = model.inference(
            chunk, 
            language, 
            gpt_cond_latent, 
            speaker_embedding, 
            temperature=temperature, 
            length_penalty=length_penalty, 
            num_beams=num_beams, 
            repetition_penalty=repetition_penalty, 
            top_k=top_k, 
            top_p=top_p,
            enable_text_splitting=enable_text_splitting,
            do_sample=do_sample
        )
        audio_tensor = torch.tensor(out["wav"])

        if post_process_audio:
            # Apply noise reduction and board processing only if post_process_audio is True
            reduced_noise = noisereduce.reduce_noise(y=audio_tensor.numpy(), sr=24000, stationary=True, prop_decrease=0.75)
            processed_audio = board(reduced_noise, 24000)
        else:
            # If post processing is not enabled, use the original audio tensor
            processed_audio = audio_tensor.numpy()

        if save_individual_audio:
            individual_output_path = os.path.join(subfolder_path, f"{timestamp}_part_{counter}.wav")
            torchaudio.save(individual_output_path, torch.tensor(processed_audio).unsqueeze(0), 24000)
        else:
            combined_audio.append(torch.tensor(processed_audio))

        counter += 1

    if not save_individual_audio:
        combined_audio_tensor = torch.cat(combined_audio, dim=0)
        output_path = os.path.join(subfolder_path, f"{timestamp}.wav")
        torchaudio.save(output_path, combined_audio_tensor.unsqueeze(0), 24000)
    
    
    folder_name = "temp_gg"

    if not os.path.exists(folder_name):
        os.mkdir(folder_name)

    if save_as_mp4:
        # Prepare parameters for the visualize function
        audio_path = output_path  # Path to the WAV file
        mp4_path = output_path.replace(".wav", ".mp4")  # Output MP4 file path

        # Call the visualize function from seewav
        visualize(audio=audio_path,
                    tmp=folder_name,  # File path of the temporary directory
                    out=mp4_path,  # Output MP4 file path
                    rate=24,  # Video framerate
                    size=(1280, 720),
fg_color = (0.5, 0.4, 0.6),  # Dark gray for the waveform
fg_color2 = (0.8, 0.2, 0.4),  # Purple-ish color for highlights
bg_color = (0.1, 0.1, 0.1))        # White background)  # Video size

        # Check if the MP4 file was successfully created
        if not os.path.exists(mp4_path):
            mp4_path = None
    else:
        mp4_path = None

    if os.path.exists(folder_name):
        shutil.rmtree(folder_name)

    end_time = datetime.datetime.now()
    duration = (end_time - start_time).total_seconds()
    per_word_duration = duration / len(text.split())
    return output_path, duration, per_word_duration * 1000, mp4_path, ""


def list_large_models(folder, min_size_gb):
    large_files = []
    min_size_bytes = min_size_gb * (1024 ** 3)
    for root, dirs, files in os.walk(folder):
        for file in files:
            if file.endswith('.pth'):
                file_path = os.path.join(root, file)
                if os.path.getsize(file_path) > min_size_bytes:
                    large_files.append(file_path)
    return large_files

reference_audio_folder = "reference_audios"
if not os.path.exists(reference_audio_folder):
    os.makedirs(reference_audio_folder)
    reference_audios = ["No available reference audios"]
    reference_audio_warning = "Warning: 'reference_audio' folder is empty. Please add audio files."
else:
    reference_audios = os.listdir(reference_audio_folder)
    reference_audio_warning = ""

train_folder = "TTS_models"
if not os.path.exists(train_folder):
    os.makedirs(train_folder)

min_size_gb = 1
model_folders = list_large_models(train_folder, min_size_gb)
model_warning = ""
if not model_folders:
    model_warning = "Warning: No large model files found in 'train' directory."

languages = ["en","es","fr","de","it","pt","pl","tr","ru","nl","cs","ar","zh-cn","hu","ko","ja"]

with gr.Blocks() as iface:
    with gr.Row():
        submit_button = gr.Button("Submit")
        clear_button = gr.Button("Clear")

    with gr.Row():
        with gr.Column():
            model_dropdown = gr.Dropdown(label="Model - scans TTS_models folder", choices=model_folders)
            reference_audio_dropdown = gr.Dropdown(label="Reference Audio - scans reference_audios folder", choices=reference_audios)
            reference_audio_upload = gr.Audio(label="Or Upload Reference Audio", type="filepath")
            text_box = gr.Textbox(label="Text to Transcribe", lines=5, placeholder="Enter text here...")
            language_dropdown = gr.Dropdown(label="Language", choices=languages, value="en")
            temperature_slider = gr.Slider(label="Temperature", minimum=0.01, maximum=10.0, step=0.01, value=0.75)
            length_penalty_slider = gr.Slider(label="Length Penalty", minimum=0.1, maximum=10.0, step=0.1, value=1.0)
            num_beams_slider = gr.Slider(label="Num Beams", minimum=1, maximum=10, step=1, value=1)
            repetition_penalty_slider = gr.Slider(label="Repetition Penalty", minimum=0.1, maximum=20.0, step=0.1, value=10.0)
            top_k_slider = gr.Slider(label="Top K", minimum=1, maximum=200, step=1, value=50)
            top_p_slider = gr.Slider(label="Top P", minimum=0.01, maximum=10.0, step=0.01, value=0.85)
            do_sample_checkbox = gr.Checkbox(label="Do Sample", value=True)
            gpt_cond_len_slider = gr.Slider(label="GPT Condition Length", minimum=1, maximum=600, step=1, value=30)
            gpt_cond_chunk_len_slider = gr.Slider(label="GPT Condition Chunk Length", minimum=1, maximum=20, step=1, value=6)
            max_ref_len_slider = gr.Slider(label="Max Reference Length", minimum=1, maximum=300, step=1, value=10)
            sound_norm_refs_checkbox = gr.Checkbox(label="Normalize Sound References", value=True)
            post_process_audio_checkbox = gr.Checkbox(label="Post Process Audio to Improve Quality", value=True)
            save_as_mp4_checkbox = gr.Checkbox(label="Save also as MP4", value=False)
            save_individual_audio_checkbox = gr.Checkbox(label="Save Individual Audio Files", value=False)

        with gr.Column():
            audio_output = gr.Audio(label="Generated Audio")
            generation_time_output = gr.Number(label="Generation Time (seconds)")
            per_word_duration_output = gr.Number(label="Per Word Duration (ms)")
            video_output = gr.Video(label="Generated Video")
            warnings_output = gr.Textbox(label="Warnings", value=f"{model_warning}\n{reference_audio_warning}")

    def process_inputs(model_path, reference_audio_dropdown, reference_audio_upload, text, language, temperature, length_penalty, num_beams, repetition_penalty, top_k, top_p, do_sample, gpt_cond_len, gpt_cond_chunk_len, max_ref_len, sound_norm_refs, save_individual_audio, post_process_audio, save_as_mp4):
        return inference(
			model_path, reference_audio_dropdown, reference_audio_upload, text, language, temperature, length_penalty, num_beams, repetition_penalty, top_k, top_p, do_sample, gpt_cond_len, gpt_cond_chunk_len, max_ref_len, sound_norm_refs, save_individual_audio, post_process_audio, save_as_mp4
		)

    submit_button.click(
        process_inputs, 
        inputs=[
            model_dropdown, reference_audio_dropdown, reference_audio_upload, text_box, language_dropdown, 
            temperature_slider, length_penalty_slider, num_beams_slider, repetition_penalty_slider, top_k_slider, 
            top_p_slider, do_sample_checkbox, gpt_cond_len_slider, gpt_cond_chunk_len_slider, max_ref_len_slider, 
            sound_norm_refs_checkbox, save_individual_audio_checkbox, post_process_audio_checkbox, save_as_mp4_checkbox
        ],
        outputs=[audio_output, generation_time_output, per_word_duration_output,video_output, warnings_output ]
    )

    def clear_inputs():
        text_box.value = ""

    clear_button.click(clear_inputs)

iface.launch(share=False, inbrowser=True)